function moduloRecovery()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Code for the modulo denoising
% M. Cucuringu, H. Tyagi,
% "Provably robust estimation of modulo 1 samples of a smooth function with applications to phase unwrapping" 
% (2018) accepted to Journal of Machine Learning Research (JMLR)
% (arXiv)  https://arxiv.org/abs/1803.03669
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


n = 500;   %% number of samples
noiseModel = 'Gaussian';   %% the noise model is fixed to be Gaussian

lambda = 0.1;  %% regularization parameter
sigma = 0.05;  %% noise level 
k = 3;         %% radius connectivity 



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all
% clf;
rng default
warning('off','all')
addpath(genpath('helpers'));

%% The function
ff = @(x) 4 + 4.*x .* cos(2*pi*x) .^2 - 2.*sin(2*pi*x).^2 + 0.7;   

%% Interval [a,b] fixed to [0,1]
a = 0;    b = 1; 
PARS.f = ff;

%% Uniform grid of n samples
eps = 0.0001;
grid_a = a + eps;
grid_b = b - eps;
x = grid_a:( (grid_b-grid_a) /(n-1)):grid_b; % uniform grid of n samples
if ( n ~= length(x) )
    x = [ x grid_b ];
end

f_clean = ff(x);
f_mod1_clean = mod(f_clean,1);  % Clean Mod 1 samples of function
q_clean = f_clean - f_mod1_clean;
n = length(x);

PARS.noiseModel = noiseModel;

f_noise = f_clean + sigma * randn(1,n);
f_mod1_noise = mod(f_noise,1);  % (Noisy) Mod 1 samples of function
labelNoise = '\sigma';

f_mod1_noise = mod(f_noise,1);  % Clean Mod 1 samples of function
y = f_mod1_noise;

n = length(y);
abl = 1:(n-1); % all but last
deltas = y(abl) - y(2:n);


%% Plots
figure(1);
plot(x,f_clean,'-b','LineWidth', 2); hold on;
legList{ 1 } = 'clean function (f)';   nrPlots = 1;

if sigma > 0
    plot(x,f_noise,'-g','LineWidth', 2); hold on;
    nrPlots = nrPlots + 1;
    legList{ nrPlots } = 'noisy function';
end


stem(x,y,'-k','MarkerSize', 4, 'LineWidth',1 );
nrPlots = nrPlots + 1;
legList{ nrPlots } = 'noisy f mod 1 (y)';
hold on;    plot(x,q_clean, '-r','LineWidth', 2);
nrPlots = nrPlots + 1;
legList{ nrPlots } = 'true quotient (q)';
hold on
stem(x(abl), deltas, 'm.', 'MarkerSize', 5, 'LineWidth',2)
nrPlots = nrPlots + 1;
legList{ nrPlots } =  '\delta_i = y_i - y_{i+1}';


%% Run QCQP + OLS
PARS.dim = 1;
PARS.n = n;
PARS.k = k;
to_plot = 0;
x_grid = x;
DATA = [];

PARS.method = 'QCQP';

[ f_QCQP, f_mod1_QCQP, err_f_QCQP, err_f_mod1_QCQP ] = ...
                        qcqp_solver( f_clean, f_mod1_clean, f_noise, f_mod1_noise, ...
                                            k, sigma, x_grid, DATA ,PARS, to_plot, lambda);
qest = f_QCQP - f_mod1_QCQP;
qest = optimalShift(qest, q_clean);
errorsQ = getErrors(qest, q_clean);
qest = qest + 2;  % manually shift for better vizualization

%% Plot results:
hold on;  plot(x,qest ,'--r','LineWidth', 2); 
nrPlots = nrPlots + 1;   legList{ nrPlots } = 'recovered quotient';

f_QCQP = f_QCQP+2;   % manually shift for better vizualization
hold on;  plot(x,f_QCQP ,'-c','LineWidth', 2); 
nrPlots = nrPlots + 1;   legList{ nrPlots } = 'recovered f';

PARS.myFont=24;
PARS.myFontTitle = 29;

legend(legList, 'Location','northwest');

xlabel('x'); axis tight;
set(gca,'FontSize',PARS.myFont);

h_xlabel = get(gca,'XLabel');   h_ylabel = get(gca,'YLabel');
set(h_xlabel,'FontSize',PARS.myFont); set(h_ylabel,'FontSize',PARS.myFont);

FigHandle = figure(1);  set(FigHandle, 'Position', [200, 400, 950, 750]);
set(gcf, 'PaperPositionMode','auto');


%% Recovery errors:
disp('Recovery errors: f  &  f mod 1');
disp([ err_f_QCQP, err_f_mod1_QCQP ]);

end